#!/usr/bin/python3
import os
from .bar import Bar


class Device(object):
    """ Create a Device object for a PCIe device.

    :param str pciid: PCI bus ID as string, e.g. "0000:03:00.0"

    """

    __base = "/sys/bus/pci/devices/"

    def __init__(self, pciid: str):
        self.__base = os.path.join(self.__base, str(pciid))
        if not os.access(self.__base, os.F_OK):
            raise ValueError("Device not found: %s" % (self.__base))
        self.bar = [None] * 6
        for barnum in range(0, 6):
            resfile = os.path.join(self.__base, "resource%d" % (barnum))
            if os.access(resfile, os.F_OK):
                self.bar[barnum] = Bar(resfile)

    def __get_attr__(self, attr: str, attr_type: type):
        """ Read a sysfs attribute and convert the received string to the
        given attribute type.

        :param str attr: attribute name
        :param type attr_type: data type the attribute is casted to.
        :returns: attribute value
        :rtype: attr_type

        """
        path = os.path.join(self.__base, attr)
        if not os.access(path, os.F_OK):
            raise ValueError("Cannot read attribute %s" % (attr))
        with open(path) as f:
            val = f.read()[:-1]  # strip newline
            if attr_type is int:
                return attr_type(val, 0)
            else:
                return attr_type(val)

    def vendor(self) -> int:
        """ Get the PCI vendor ID.

        :returns: PCI vendor ID
        :rtype: int
        """
        return self.__get_attr__("vendor", int)

    def device(self) -> int:
        """ Get the PCI device ID.

        :returns: PCI device ID
        :rtype: int
        """
        return self.__get_attr__("device", int)

    def revision(self) -> int:
        """ Get the PCI revision Number.

        :returns: PCI revision Number
        :rtype: int
        """
        return self.__get_attr__("revision", int)

    def subsystem_vendor(self) -> int:
        """ Get the PCI subsystem vendor ID.

        :returns: PCI subsystem vendor ID
        :rtype: int
        """
        return self.__get_attr__("subsystem_vendor", int)

    def subsystem_device(self) -> int:
        """ Get the PCI subsystem device ID.

        :returns: PCI subsystem device ID
        :rtype: int
        """
        return self.__get_attr__("subsystem_device", int)



from setuptools import setup, find_packages

VERSION = '0.1.1'
DESCRIPTION = 'Simple Python module to access PCIe devices'
with open("README.md", "r", encoding="utf-8") as fh:
    LONG_DESCRIPTION = fh.read()

# Setting up
setup(
    name="pyPCIe",
    version=VERSION,
    author="Heiko Engel",
    author_email="<heikoengel@users.noreply.github.com>",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    url="https://github.com/heikoengel/pyPCIe",
    install_requires=[],
    keywords=['python', 'pcie'],
    classifiers= [
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
    ]
)



#!/usr/bin/python3
import os
from mmap import mmap, PROT_READ, PROT_WRITE, PAGESIZE
from struct import pack, unpack


class Bar(object):
    """ Create a Bar instance that ``mmap()`` s a PCIe device BAR.

    :param str filename: sysfs filename for the corresponding resourceX file.

    """

    def __init__(self, filename: str):
        self.__map = None
        self.__stat = os.stat(filename)
        fd = os.open(filename, os.O_RDWR)
        self.__map = mmap(fd, 0, prot=PROT_READ | PROT_WRITE)
        os.close(fd)

    def __del__(self):
        if self.__map is not None:
            self.__map.close()

    def __check_offset(self, offset: int):
        """ Check if the given offset is properly DW-aligned and the access
            falls within the BAR size.

        """
        if offset & 0x3:
            raise ValueError("unaligned access to offset 0x%x" % (offset))
        if offset + 3 > self.size:
            raise ValueError("offset (0x%x) exceeds BAR size (0x%x)" %
                             (offset, self.size))

    def read(self, offset: int):
        """ Read a 32 bit / double word value from offset.

        :param int offset: BAR byte offset to read from.
        :returns: Double word read from the given BAR offset.
        :rtype: double word / 32 bit unsigned long / int

        """
        self.__check_offset(offset)
        reg = self.__map[offset:offset+4]
        return unpack("<L", reg)[0]

    def write(self, offset: int, data: int):
        """ Write a 32 bit / double word value to offset.

        :param int offset: BAR byte offset to write to.
        :param int data: double word to write to the given BAR offset.
        """
        self.__check_offset(offset)
        self.__map.seek(offset)
        reg = pack("<L", data)
        # write to map. no ret. check: ValueError/TypeError is raised on error
        self.__map.write(reg)
        # Flush current page for immediate update.
        page_offset = offset & (~(PAGESIZE - 1) & 0xffffffff)
        self.__map.flush(page_offset, PAGESIZE)
        # TODO: check return value, only for >=Python3.8

    @property
    def size(self):
        """
        Get the size of the BAR.

        :returns: BAR size in bytes.
        :rtype: int
        """
        return self.__stat.st_size
